local function PatchMissingTables(strings, startdepth, sep)
  if not strings then return end
  startdepth = startdepth or 1
  for k, v in pairs(strings) do
    local keys = {undotted(k)}
    for i = 1, startdepth do table.remove(keys, 1) end
    local leaf = table.remove(keys, #keys)
    local root = STRINGS
    for i, node in ipairs(keys) do
      local oldval = root[node]
      if type(oldval) ~= "table" then root[node] = {} end
      root = root[node]
    end
    if type(tonumber(leaf)) == "number" then
      root[tonumber(leaf)] = v -- patched number circumstances
    else
      -- root[leaf] = sep and separator(v) or v
      root[leaf] = v
    end
  end
end
local function MergeTranslationFromLUA(dir, filename, lang, sep)
  local content = require(dir .. filename .. "_" .. lang)
  PatchMissingTables(content, 1, sep)
end
local function MergeTable(table1, table2)
  if not table1 or not table2 then return end
  for key, value in pairs(table2) do
    if type(tonumber(key)) == "number" then key = tonumber(key) end
    if table1[key] == nil then
      if type(value) == "table" then
        table1[key] = {}
        MergeTable(table1[key], value)
      else
        table1[key] = value
      end
    else
      if type(value) == "table" then
        if type(table1[key]) ~= "table" then
          local val = table1[key]
          print("automatically joined old value to generic", key, val)
          table1[key] = {GENERIC = val}
        end
        MergeTable(table1[key], value)
      else
        table1[key] = value
      end
    end
  end
end

local function MergeTranslationFromJSON(dir, filename, lang)
  local f = MakeFile(softresolvefilepath(dir .. filename .. "_" .. lang .. ".json"))
  f.persist = false
  local data = f:read()
  if data then
    local success, result = pcall(json.decode, data)
    if success then
      MergeTable(STRINGS, result)
    else
      print("[TranslateLanguages] file: " .. filename .. "cannot be parsed")
    end
  else
    print("[TranslateLanguages] file: " .. filename .. "cannot be opened")
  end
end
  local jsonfiles = {
    "MERMGUARD",
    "MERMHOUSE_CRAFTED",
    "MERMKING",
    "MERMNAMES",
    "MERMTHRONE",
    "MERMWATCHTOWER",
    "MERM_BATTLECRY",
    "MERM_GUARD_BATTLECRY",
    "MERM_KING",
    "MERM_PRINCE",
    "MERM_TALK",
    "WURT"
  }
  local charfiles = {}
  local sfiles = {"wurt"}
  local fbase = "languages/"
  local jbase = "scripts/languages/"
  local supported = {zh = true, en = false}
  local lang = LanguageTranslator.defaultlang
  lang = supported[lang] ~= nil and lang or "en"
  
  
  --here is an insertion
  if lang=="zh" then
  STRINGS.NAMES.WURT = "沃特"
    STRINGS.CHARACTERS.WURT = STRINGS.CHARACTERS.WURT or {}
    STRINGS.CHARACTER_NAMES.wurt = "沃特"
    STRINGS.CHARACTER_TITLES.wurt = "浮浪噗"
    STRINGS.CHARACTER_DESCRIPTIONS.wurt = [[*一个鱼人小女孩
*想让自己的族人重回昔日的辉煌
*特别喜欢看书，因为里面有漂亮的图片]]
    STRINGS.CHARACTER_QUOTES.wurt = "“也许别人会对沼泽褒贬不一，我却乐此不疲。”"
else
STRINGS.NAMES.WURT = "Wurt"
STRINGS.CHARACTERS.WURT = STRINGS.CHARACTERS.WURT or {}
STRINGS.CHARACTER_NAMES.wurt = "Wurt"
STRINGS.CHARACTER_TITLES.wurt = "Florp"
STRINGS.CHARACTER_DESCRIPTIONS.wurt = [[*A little Merm girl.
*Make Merms Great Again.
*Fanscinated by pretty books.]]
STRINGS.CHARACTER_QUOTES.wurt = "\"Though some folks might think the swamp's no place to be, I'd kinda disagree.\""
end
  
  
  
  
  for i, v in ipairs(jsonfiles) do MergeTranslationFromJSON(jbase, v, lang) end
  for _, name in ipairs(charfiles) do
    local uname = string.upper(name)
    if STRINGS.CHARACTERS[uname] then -- DLC characters may not exist
      local characterspeech = require(fbase .. name .. "_" .. lang)
      local category = "DESCRIBE" -- note:hack to shrink data
      for _, tbl in pairs(characterspeech) do
        if not STRINGS.CHARACTERS[uname][category] then STRINGS.CHARACTERS[uname][category] = {} end
        MergeTable(STRINGS.CHARACTERS[uname][category], tbl)
      end
      package.loaded[fbase .. name .. "_" .. lang] = nil
    end
  end
  for i, name in ipairs(sfiles) do
    local require_path = fbase .. "speech_" .. name .. "_" .. lang
    local file_path = "scripts/" .. require_path .. ".lua"
    if softresolvefilepath(file_path) then
      local characterspeech = require(require_path)
      MergeTable(STRINGS, characterspeech)
      package.loaded[fbase .. name .. "_" .. lang] = nil
    end
  end
collectgarbage("collect")
