local BrainCommon = require("brains/braincommon")
--------------------------------------------------------------------------
-- Actions: MINE, CHOP, HACK

local MINE_TAGS = {"MINE_workable"}
local MINE_CANT_TAGS = {"carnivalgame_part", "event_trigger"}
local CHOP_TAGS = {"CHOP_workable"}
local CHOP_CANT_TAGS = {"carnivalgame_part", "event_trigger"}

local function IsDeciduousTreeMonster(guy) return guy.monster and guy.prefab == "deciduoustree" end

local function FindDeciduousTreeMonster(inst, finddist)
  return FindEntity(inst, finddist / 3, IsDeciduousTreeMonster, CHOP_TAGS)
end

AssistLeaderDefaults = {
  MINE = {
    Starter = function(inst, leaderdist, finddist)
      return inst.components.follower.leader ~= nil and inst.components.follower.leader.sg ~= nil and
              inst.components.follower.leader.sg:HasStateTag("mining")
    end,
    KeepGoing = function(inst, leaderdist, finddist)
      return inst.components.follower.leader ~= nil and inst:IsNear(inst.components.follower.leader, leaderdist)
    end,
    FindNew = function(inst, leaderdist, finddist)
      local target = FindEntity(inst, finddist, nil, MINE_TAGS, MINE_CANT_TAGS)

      if target == nil and inst.components.follower.leader ~= nil then
        target = FindEntity(inst.components.follower.leader, finddist, nil, MINE_TAGS, MINE_CANT_TAGS)
      end

      if target ~= nil then return BufferedAction(inst, target, ACTIONS.MINE) end
    end
  },
  CHOP = {
    Starter = function(inst, finddist)
      return inst.tree_target ~= nil or
              (inst.components.follower.leader ~= nil and inst.components.follower.leader.sg ~= nil and
               inst.components.follower.leader.sg:HasStateTag("chopping")) or FindDeciduousTreeMonster(inst, finddist) ~=
              nil
    end,
    KeepGoing = function(inst, leaderdist, finddist)
      return inst.tree_target ~= nil or
              (inst.components.follower.leader ~= nil and inst:IsNear(inst.components.follower.leader, leaderdist)) or
              FindDeciduousTreeMonster(inst, finddist) ~= nil
    end,
    FindNew = function(inst, leaderdist, finddist)
      local target = FindEntity(inst, finddist, nil, CHOP_TAGS, CHOP_CANT_TAGS)

      if target == nil and inst.components.follower.leader ~= nil then
        target = FindEntity(inst.components.follower.leader, finddist, nil, CHOP_TAGS, CHOP_CANT_TAGS)
      end

      if target ~= nil then
        if inst.tree_target ~= nil then
          target = inst.tree_target
          inst.tree_target = nil
        else
          target = FindDeciduousTreeMonster(inst, finddist) or target
        end

        return BufferedAction(inst, target, ACTIONS.CHOP)
      end
    end
  },
  HACK = {
    KeepGoing = function(inst, leaderdist, finddist)
      return (inst.components.follower.leader ~= nil and inst:IsNear(inst.components.follower.leader, leaderdist))
    end,
    Starter = function(inst)
      return inst.components.follower.leader ~= nil and inst.components.follower.leader.sg ~= nil and
              inst.components.follower.leader.sg:HasStateTag("hacking")
    end,
    FindNew = function(inst, leaderdist, finddist)
      local target = FindEntity(inst, finddist, function(inst)
        return inst.components.hackable and inst.components.hackable:CanBeHacked()
      end)

      if target == nil and inst.components.follower.leader ~= nil then
        target = FindEntity(inst.components.follower.leader, finddist, function(inst)
          return inst.components.hackable and inst.components.hackable:CanBeHacked()
        end)
      end
      if target ~= nil then return BufferedAction(inst, target, ACTIONS.HACK) end
    end
  }
}
-- Mod support access.
BrainCommon.AssistLeaderDefaults = AssistLeaderDefaults

-- NOTES(JBK): This helps followers do a task once they see the leader is doing an act.
--            Since actions are very context sensitive, there are defaults above to help clarify context.
local function NodeAssistLeaderDoAction(self, parameters)
  local action = parameters.action
  local defaults = AssistLeaderDefaults[action]

  local starter = parameters.starter or defaults.Starter
  local keepgoing = parameters.keepgoing or defaults.KeepGoing
  local finder = parameters.finder or defaults.FindNew

  local keepgoing_leaderdist = parameters.keepgoing_leaderdist or TUNING.FOLLOWER_HELP_LEADERDIST
  local finder_finddist = parameters.finder_finddist or TUNING.FOLLOWER_HELP_FINDDIST

  local function ifnode() return ACTIONS[action] ~= nil and starter(self.inst, keepgoing_leaderdist, finder_finddist) end
  local function whilenode() return keepgoing(self.inst, keepgoing_leaderdist, finder_finddist) end
  local function findnode() return finder(self.inst, keepgoing_leaderdist, finder_finddist) end
  local looper
  if parameters.chatterstring then
    looper = LoopNode{ConditionNode(whilenode),
                      ChattyNode_DST(self.inst, parameters.chatterstring, DoAction(self.inst, findnode))}
  else
    looper = LoopNode{ConditionNode(whilenode), DoAction(self.inst, findnode)}
  end

  return IfThenDoWhileNode(ifnode, whilenode, action, looper)
end

BrainCommon.NodeAssistLeaderDoAction = NodeAssistLeaderDoAction
--------------------------------------------------------------------------
function GetClosestInstWithTag(tag, inst, radius)
  local x, y, z = inst.Transform:GetWorldPosition()
  local ents = TheSim:FindEntities(x, y, z, radius, type(tag) == "string" and {tag} or tag)
  return ents[1] ~= inst and ents[1] or ents[2]
end
local function ShouldTriggerPanic(inst)
  return (inst.components.health ~= nil and inst.components.health.takingfiredamage) or
          (inst.components.hauntable ~= nil and inst.components.hauntable.panic)
end
--------------------------------------------------------------------------
BrainCommon.ShouldTriggerPanic = ShouldTriggerPanic
BrainCommon.PanicTrigger = function(inst)
  return WhileNode(function() return ShouldTriggerPanic(inst) end, "PanicTrigger", Panic(inst))
end

--------------------------------------------------------------------------
function IfThenDoWhileNode(ifcond, whilecond, name, node)
  return ParallelNode{MultiConditionNode(ifcond, whilecond, name), node}
end
function FailNode() return BehaviourNode("Fail") end
MultiConditionNode = Class(BehaviourNode, function(self, start, continue, name)
  BehaviourNode._ctor(self, name or "Condition")
  self.start = start
  self.continue = continue
end)

function MultiConditionNode:Visit()
  if not self.running then
    self.running = self.start()
  else
    self.running = self.continue()
  end

  if self.running then
    self.status = SUCCESS
  else
    self.status = FAILED
  end
end
--------------------------------------------------------------------------
require"behaviours/follow"
local old = Follow._ctor
Follow._ctor = function(self, inst, target, min_dist, target_dist, max_dist, canrun, alwayseval, ...)
  old(self, inst, target, min_dist, target_dist, max_dist, canrun, alwayseval, ...)
  if type(target_dist) == "function" then
    self.target_dist_fn = target_dist
    self.target_dist = {}
    setmetatable(self.target_dist, {
      __mul = function(t, v2)
        local v = self.target_dist_fn(self.inst)
        local t = v * (type(v2) == "table" and v2() or v2)
        return t
      end,
      __call = function(t)
        local v = self.target_dist_fn(self.inst)
        return v
      end
    })
  end
end
