require"behaviours/wander"
require"behaviours/runaway"
require"behaviours/doaction"
require"behaviours/panic"
require"behaviours/follow"
require"behaviours/ChattyNode_DST"
require"behaviours/assistmerm"

local BrainCommon = require"brains/braincommon"
local function ShouldTriggerPanic(inst)
  return (inst.components.health ~= nil and inst.components.health.takingfiredamage) or
          (inst.components.hauntable ~= nil and inst.components.hauntable.panic)
end

BrainCommon.ShouldTriggerPanic = ShouldTriggerPanic
BrainCommon.PanicTrigger = BrainCommon.PanicTrigger or
                            function(inst)
   return WhileNode(function() return ShouldTriggerPanic(inst) end, "PanicTrigger", Panic(inst))
 end
local SEE_PLAYER_DIST = 5
local SEE_FOOD_DIST = 10
local MAX_WANDER_DIST = 15
local MAX_CHASE_TIME = 10
local MAX_CHASE_DIST = 20
local RUN_AWAY_DIST = 5
local STOP_RUN_AWAY_DIST = 8

local MIN_FOLLOW_DIST = 3
local MAX_FOLLOW_DIST = 15

local SEE_THRONE_DISTANCE = 50

local FACETIME_BASE = 2
local FACETIME_RAND = 2

local TRADE_DIST = 20

local MermBrain = Class(Brain, function(self, inst) Brain._ctor(self, inst) end)

local function GetTraderFn(inst)
  local x, y, z = inst.Transform:GetWorldPosition()
  -- local players = FindPlayersInRange(x, y, z, TRADE_DIST, true)--changed
  local players = ThePlayer:IsNear(inst, TRADE_DIST) and {ThePlayer} or {}
  for _, v in ipairs(players) do if inst.components.trader:IsTryingToTradeWithMe(v) then return v end end
end

local function KeepTraderFn(inst, target) return inst.components.trader:IsTryingToTradeWithMe(target) end

local function GetFaceTargetFn(inst)
  if inst.components.timer:TimerExists("dontfacetime") then return nil end
  local shouldface = inst.components.follower.leader or -- FindClosestPlayerToInst(inst, SEE_PLAYER_DIST, true)--changed
  (ThePlayer:IsNear(inst, SEE_PLAYER_DIST) and ThePlayer)
  if shouldface and not inst.components.timer:TimerExists("facetime") then
    inst.components.timer:StartTimer("facetime", FACETIME_BASE + math.random() * FACETIME_RAND)
  end
  return shouldface
end

local function KeepFaceTargetFn(inst, target)
  if inst.components.timer:TimerExists("dontfacetime") then return nil end
  local keepface = (inst.components.follower.leader and inst.components.follower.leader == target) or
                    (target:IsValid() and inst:IsNear(target, SEE_PLAYER_DIST))
  if not keepface then inst.components.timer:StopTimer("facetime") end
  return keepface
end

function GetClosestInstWithTag(tag, inst, radius)
  local x, y, z = inst.Transform:GetWorldPosition()
  local ents = TheSim:FindEntities(x, y, z, radius, type(tag) == "string" and {tag} or tag)
  return ents[1] ~= inst and ents[1] or ents[2]
end
local function EatFoodAction(inst)
  local target = nil
  if inst.components.inventory and inst.components.eater then
    target = inst.components.inventory:FindItem(function(item) return inst.components.eater:CanEat(item) end)
  end
  if not target then
    target = FindEntity(inst, SEE_FOOD_DIST, function(item)
      return inst.components.eater:CanEat(item) and item:IsOnValidGround() and not item:HasTag("INLIMBO")
    end)
    if target then
      -- check for scary things near the food
      local predator = GetClosestInstWithTag("scarytoprey", target, SEE_PLAYER_DIST)
      if predator then target = nil end
    end
  end
  if target then
    local act = BufferedAction(inst, target, ACTIONS.EAT)
    act.validfn = function()
      return not (target.components.inventoryitem and target.components.inventoryitem.owner and
              target.components.inventoryitem.owner ~= inst)
    end
    return act
  end
end

-----------------------------------------------------------------------------------------------
-- Merm king things
local function IsThroneValid(inst)
  if TheWorld.components.mermkingmanager then
    local throne = TheWorld.components.mermkingmanager:GetThrone(inst)
    return throne ~= nil and throne:IsValid() and
            not (throne.components.burnable ~= nil and throne.components.burnable:IsBurning()) and
            not throne:HasTag("burnt") and TheWorld.components.mermkingmanager:ShouldGoToThrone(inst, throne)
  end

  return false
end

local GOTOTHRONE_TAGS = {"mermthrone"}
local function ShouldGoToThrone(inst)
  if TheWorld.components.mermkingmanager then
    local throne = TheWorld.components.mermkingmanager:GetThrone(inst)
    if throne == nil then throne = FindEntity(inst, SEE_THRONE_DISTANCE, nil, GOTOTHRONE_TAGS) end

    return throne and TheWorld.components.mermkingmanager:ShouldGoToThrone(inst, throne)
  end

  return false
end

local function GetThronePosition(inst)
  if TheWorld.components.mermkingmanager then
    local throne = TheWorld.components.mermkingmanager:GetThrone(inst)
    if throne then return throne:GetPosition() end
  end
end

-----------------------------------------------------------------------------------------------

local function GoHomeAction(inst)
  if inst.components.combat.target ~= nil then return end
  local home = inst.components.homeseeker ~= nil and inst.components.homeseeker.home or nil
  return home ~= nil and home:IsValid() and
          not (home.components.burnable ~= nil and home.components.burnable:IsBurning()) and not home:HasTag("burnt") and
          BufferedAction(inst, home, ACTIONS.GOHOME) or nil
end

local function ShouldGoHome(inst)
  if not TheWorld.state.isday or (inst.components.follower ~= nil and inst.components.follower.leader ~= nil) then
    return false
  end

  -- one merm should stay outside
  local home = inst.components.homeseeker ~= nil and inst.components.homeseeker.home or nil
  return home == nil or home.components.childspawner == nil or home.components.childspawner:CountChildrenOutside() > 1
end

local function IsHomeOnFire(inst)
  return inst.components.homeseeker and inst.components.homeseeker.home and
          inst.components.homeseeker.home.components.burnable and
          inst.components.homeseeker.home.components.burnable:IsBurning()
end

local function GetNoLeaderHomePos(inst)
  if inst.components.follower and inst.components.follower.leader ~= nil then return nil end

  return inst.components.knownlocations:GetLocation("home")
end

local function CurrentContestTarget(inst)
  local stage = inst.npc_stage
  if stage.current_contest_target then
    return stage.current_contest_target
  else
    return stage
  end
end

local function MarkPost(inst)
  -- deleted--if inst.yotb_post_to_mark ~= nil then return BufferedAction(inst, inst.yotb_post_to_mark, ACTIONS.MARK) end
end

local function CollctPrize(inst)
  if inst.yotb_prize_to_collect ~= nil then
    local x, y, z = inst.yotb_prize_to_collect.Transform:GetWorldPosition()
    if y < 0.1 and y > -0.1 and not inst.yotb_prize_to_collect:HasTag("INLIMBO") then
      return BufferedAction(inst, inst.yotb_prize_to_collect, ACTIONS.PICKUP)
    end
  end
end

local function TargetFollowDistFn(inst)
  local loyalty = inst.components.follower ~= nil and inst.components.follower:GetLoyaltyPercent() or 0.5
  local boatmod = inst:GetCurrentPlatform() ~= nil and 0.2 or 1.0
  -- As loyalty rises the further out the follower will potentially stay back.
  -- If on a boat lower max range.
  -- Randomize the range from min to max with the bias modulation.
  return (MAX_FOLLOW_DIST - MIN_FOLLOW_DIST) * (1.0 - loyalty) * boatmod * math.random() + MIN_FOLLOW_DIST
end

function MermBrain:OnStart()
  local in_contest = WhileNode(function() return self.inst:HasTag("NPC_contestant") end, "In contest", PriorityNode(
   { -- IfNode(function() return self.inst.yotb_post_to_mark end, "mark post",   DoAction(self.inst, CollctPrize, "collect prize", true), DoAction(self.inst, MarkPost, "mark post", true), )
   WhileNode(
    function() return self.inst.components.timer and self.inst.components.timer:TimerExists("contest_panic") end,
    "Panic Contest", ChattyNode_DST(self.inst, "MERM_TALK_CONTEST_PANIC", Panic(self.inst))),
   ChattyNode_DST(self.inst, "MERM_TALK_CONTEST_OOOH", FaceEntity(self.inst, CurrentContestTarget, CurrentContestTarget), 5,
    15)}, 0.1))

  local root = PriorityNode({IfNode(function()
    return TheWorld.components.mermkingmanager and TheWorld.components.mermkingmanager.king
  end, "panic with king", BrainCommon.PanicWhenScared(self.inst, .25, "MERM_TALK_PANICBOSS_KING")),
                             IfNode(
   function() return not TheWorld.components.mermkingmanager or not TheWorld.components.mermkingmanager.king end,
   "panic with no king", BrainCommon.PanicWhenScared(self.inst, .25, "MERM_TALK_PANICBOSS")),
                             BrainCommon.PanicTrigger(self.inst), WhileNode(
   function() return self.inst.components.combat.target == nil or not self.inst.components.combat:InCooldown() end,
   "AttackMomentarily", ChaseAndAttack(self.inst, SpringCombatMod(MAX_CHASE_TIME), SpringCombatMod(MAX_CHASE_DIST))),
                             WhileNode(
   function() return self.inst.components.combat.target ~= nil and self.inst.components.combat:InCooldown() end,
   "Dodge",
   RunAway(self.inst, function() return self.inst.components.combat.target end, RUN_AWAY_DIST, STOP_RUN_AWAY_DIST)),

                             in_contest,

                             ChattyNode_DST(self.inst, "MERM_TALK_FIND_FOOD", DoAction(self.inst, EatFoodAction, "Eat Food")), -- NOTES(JBK): Leave this above throne task so the Merm eats the food given.

                             WhileNode(
   function() return ShouldGoToThrone(self.inst) and self.inst.components.combat.target == nil end, "ShouldGoToThrone",
   PriorityNode({Leash(self.inst, GetThronePosition, 0.2, 0.2, true),
                 IfNode(function() return IsThroneValid(self.inst) end, "IsThroneValid",
    ActionNode(function() self.inst:PushEvent("onarrivedatthrone") end))}, .25)),
                             BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "HACK", -- Required.
    chatterstring = "MERM_TALK_HELP_HACK_BUSH"
  }), BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "CHOP", -- Required.
    chatterstring = "MERM_TALK_HELP_CHOP_WOOD"
  }), BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "MINE", -- Required.
    chatterstring = "MERM_TALK_HELP_MINE_ROCK"
  }), ChattyNode_DST(self.inst, "MERM_TALK_FIND_FOOD", -- TODO(JBK): MERM_TALK_ATTEMPT_TRADE
  FaceEntity(self.inst, GetTraderFn, KeepTraderFn)), ChattyNode_DST(self.inst, "MERM_TALK_FOLLOWWILSON",
   Follow(self.inst, function() return self.inst.components.follower.leader end, MIN_FOLLOW_DIST, TargetFollowDistFn,
    MAX_FOLLOW_DIST, nil, true)),
                             IfNode(function() return self.inst.components.follower.leader ~= nil end, "HasLeader",
   ChattyNode_DST(self.inst, "MERM_TALK_FOLLOWWILSON", FaceEntity(self.inst, GetFaceTargetFn, KeepFaceTargetFn))),
                             WhileNode(function() return IsHomeOnFire(self.inst) end, "HomeOnFire", Panic(self.inst)),
                             WhileNode(function() return ShouldGoHome(self.inst) end, "ShouldGoHome",
   DoAction(self.inst, GoHomeAction, "Go Home", true)), FaceEntity(self.inst, GetFaceTargetFn, KeepFaceTargetFn),
                             Wander(self.inst, GetNoLeaderHomePos, MAX_WANDER_DIST)}, .25)

  self.bt = BT(self.inst, root)
end

return MermBrain
