require"behaviours/wander"
require"behaviours/runaway"
require"behaviours/doaction"
require"behaviours/follow"
require"behaviours/ChattyNode_DST"
require"behaviours/assistmerm"
local BrainCommon = require"brains/braincommon"

local SEE_PLAYER_DIST = 5
local SEE_FOOD_DIST = 5
local MAX_WANDER_DIST = 15
local MAX_CHASE_TIME = 25
local MAX_CHASE_DIST = 40
local RUN_AWAY_DIST = 5
local STOP_RUN_AWAY_DIST = 8

local MIN_FOLLOW_DIST = 3
local MAX_FOLLOW_DIST = 15

local FACETIME_BASE = 2
local FACETIME_RAND = 2

local TRADE_DIST = 20

local MermBrain = Class(Brain, function(self, inst) Brain._ctor(self, inst) end)

local function GetTraderFn(inst)
  -- changed
  ----local x, y, z = inst.Transform:GetWorldPosition()
  ----local players = FindPlayersInRange(x, y, z, TRADE_DIST, true)
  local players = {ThePlayer}
  for _, v in ipairs(players) do if inst.components.trader:IsTryingToTradeWithMe(v) then return v end end
end

local function KeepTraderFn(inst, target) return inst.components.trader:IsTryingToTradeWithMe(target) end

local function GetFaceTargetFn(inst)
  if inst.components.timer:TimerExists("dontfacetime") then return nil end
  local shouldface = inst.components.follower.leader or FindEntity(inst,SEE_PLAYER_DIST,function(guy)return guy==ThePlayer end) -- FindClosestPlayerToInst(inst, SEE_PLAYER_DIST, true)--changed
  if shouldface and not inst.components.timer:TimerExists("facetime") then
    inst.components.timer:StartTimer("facetime", FACETIME_BASE + math.random() * FACETIME_RAND)
  end
  return shouldface
end

local function KeepFaceTargetFn(inst, target)
  if inst.components.timer:TimerExists("dontfacetime") then return nil end
  local keepface = (inst.components.follower.leader and inst.components.follower.leader == target) or
                    (target:IsValid() and inst:IsNear(target, SEE_PLAYER_DIST))
  if not keepface then inst.components.timer:StopTimer("facetime") end
  return keepface
end

------------------------------------------------------------------------------
local EATFOOD_MUST_TAGS = {"edible_VEGGIE"}
local EATFOOD_CANOT_TAGS = {"INLIMBO"}
local SCARY_TAGS = {"scarytoprey"}

local function EatFoodAction(inst)
  local target = nil
  if inst.components.inventory ~= nil and inst.components.eater ~= nil then
    target = inst.components.inventory:FindItem(function(item) return inst.components.eater:CanEat(item) end)
  end
  if target == nil and inst.components.follower.leader == nil then
    target = FindEntity(inst, SEE_FOOD_DIST, function(item) return inst.components.eater:CanEat(item) end,
     EATFOOD_MUST_TAGS, EATFOOD_CANOT_TAGS)
    -- check for scary things near the food
    if target ~= nil and
     (GetClosestInstWithTag(SCARY_TAGS, target, SEE_PLAYER_DIST) ~= nil or not target:IsOnValidGround()) then -- NOTE this ValidGround check should be removed if merms start swimming
      target = nil
    end
  end
  if target ~= nil then
    local act = BufferedAction(inst, target, ACTIONS.EAT)
    act.validfn = function()
      return target.components.inventoryitem == nil or target.components.inventoryitem.owner == nil or
              target.components.inventoryitem.owner == inst
    end
    return act
  end
end

local function GetNoLeaderHomePos(inst)
  if inst.components.follower and inst.components.follower.leader ~= nil then return nil end

  return inst.components.knownlocations:GetLocation("home")
end

local function TargetFollowDistFn(inst)
  local loyalty = inst.components.follower ~= nil and inst.components.follower:GetLoyaltyPercent() or 0.5
  local boatmod = inst:GetCurrentPlatform() ~= nil and 0.2 or 1.0
  -- As loyalty rises the further out the follower will potentially stay back.
  -- If on a boat lower max range.
  -- Randomize the range from min to max with the bias modulation.
  return (MAX_FOLLOW_DIST - MIN_FOLLOW_DIST) * (1.0 - loyalty) * boatmod * math.random() + MIN_FOLLOW_DIST
end
function MermBrain:OnStart()
  local root = PriorityNode({BrainCommon.PanicTrigger(self.inst), WhileNode(
   function() return self.inst.components.combat.target == nil or not self.inst.components.combat:InCooldown() end,
   "AttackMomentarily", ChaseAndAttack(self.inst, SpringCombatMod(MAX_CHASE_TIME), SpringCombatMod(MAX_CHASE_DIST))),
                             WhileNode(
   function() return self.inst.components.combat.target ~= nil and self.inst.components.combat:InCooldown() end,
   "Dodge",
   RunAway(self.inst, function() return self.inst.components.combat.target end, RUN_AWAY_DIST, STOP_RUN_AWAY_DIST)),

                             WhileNode(function()
    if not self.inst.king or
     (not self.inst.king:IsValid() or (self.inst.king.components.health and self.inst.king.components.health:IsDead())) then
      self.inst.return_to_king = false
      if self.inst.king then self.inst.king.OnGuardDeath(self.inst) end
      self.inst.king = nil
    end

    return self.inst.return_to_king
  end, "ShouldGoToThrone",
   PriorityNode({Leash(self.inst, function() return self.inst.king:GetPosition() end, 2, 2, true),
                 IfNode(function() return true end, "IsThroneValid", ActionNode(function()
     local fx = SpawnPrefab("merm_spawn_fx")
     if fx then fx.Transform:SetPosition(self.inst.Transform:GetWorldPosition()) end
     self.inst.SoundEmitter:PlaySound("dontstarve/characters/wurt/merm/buff") -- Splash sound
     self.inst:Remove()
   end))}, .25)), BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "HACK", -- Required.
    chatterstring = "MERM_TALK_HELP_HACK_BUSH"
  }), BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "CHOP", -- Required.
    chatterstring = "MERM_TALK_HELP_CHOP_WOOD"
  }), BrainCommon.NodeAssistLeaderDoAction(self, {
    action = "MINE", -- Required.
    chatterstring = "MERM_TALK_HELP_MINE_ROCK"
  }), ChattyNode_DST(self.inst, "MERM_TALK_FIND_FOOD", -- TODO(JBK): MERM_TALK_ATTEMPT_TRADE
  FaceEntity(self.inst, GetTraderFn, KeepTraderFn)),

                             ChattyNode_DST(self.inst, "MERM_TALK_FIND_FOOD", DoAction(self.inst, EatFoodAction, "Eat Food")),

                             ChattyNode_DST(self.inst, "MERM_TALK_FOLLOWWILSON",
   Follow(self.inst, function() return self.inst.components.follower.leader end, MIN_FOLLOW_DIST, TargetFollowDistFn,
    MAX_FOLLOW_DIST, nil, true)),

                             IfNode(function() return self.inst.components.follower.leader ~= nil end, "HasLeader",
   ChattyNode_DST(self.inst, "MERM_TALK_FOLLOWWILSON", FaceEntity(self.inst, GetFaceTargetFn, KeepFaceTargetFn))),

                             FaceEntity(self.inst, GetFaceTargetFn, KeepFaceTargetFn),
                             Wander(self.inst, GetNoLeaderHomePos, MAX_WANDER_DIST)}, .25)

  self.bt = BT(self.inst, root)
end

return MermBrain
