if not rawget(_G, "TheNet") then return nil end
local MakePlayerCharacter = require("prefabs/player_common")

local assets = {
  Asset("SCRIPT", "scripts/prefabs/player_common.lua"),
  Asset("ANIM", "anim/wurt_peruse.zip"),
  Asset("ANIM", "anim/wurt_mount_peruse.zip"),
  Asset("SOUND", "sound/wurt.fsb"),
  Asset("SOUNDPACKAGE", "sound/wurt.fev"),
  Asset("ANIM", "anim/player_idles_wurt.zip")
}

local prefabs = {"wurt_tentacle_warning"}

local start_inv = {}
for k, v in pairs(TUNING.GAMEMODE_STARTING_ITEMS) do start_inv[string.lower(k)] = v.WURT end
-- prefabs = FlattenTree({ prefabs, start_inv }, true)--no FlattenTree

local function UpdateStats(inst, maxhealth, maxhunger, maxsanity)

  if inst.health_percent == 0 then inst.health_percent = nil end
  local current_health = inst.health_percent or inst.components.health:GetPercent()
  if current_health <= 0 then
    inst:DoTaskInTime(3, UpdateStats, maxhealth, maxhunger, maxsanity)
    return
  end

  inst.health_percent = nil

  local current_hunger = inst.hunger_percent or inst.components.hunger:GetPercent()
  inst.hunger_percent = nil

  local current_sanity = inst.sanity_percent or inst.components.sanity:GetPercent()
  inst.sanity_percent = nil

  inst.components.health:SetMaxHealth(maxhealth)
  inst.components.hunger:SetMax(maxhunger)
  inst.components.sanity:SetMax(maxsanity)

  inst.components.health:SetPercent(current_health)
  inst.components.hunger:SetPercent(current_hunger)
  inst.components.sanity:SetPercent(current_sanity)
end

local function RoyalUpgrade(inst, silent)
  inst.overrideskinmode = "powerup"
  inst.overrideskinmodebuild = "wurt_stage2"

  UpdateStats(inst, TUNING.WURT_HEALTH_KINGBONUS, TUNING.WURT_HUNGER_KINGBONUS, TUNING.WURT_SANITY_KINGBONUS)

  if not silent and not inst.royal then
    inst.royal = true
    inst.components.talker:Say(GetString(inst, "ANNOUNCE_KINGCREATED"))
    inst.sg:PushEvent("powerup_wurt")
    inst.SoundEmitter:PlaySound("dontstarve/characters/wurt/transform_to")
  end
end

local function RoyalDowngrade(inst, silent)
  print("royal downgrade")
  inst.overrideskinmode = nil
  inst.overrideskinmodebuild = nil

  UpdateStats(inst, TUNING.WURT_HEALTH, TUNING.WURT_HUNGER, TUNING.WURT_SANITY)

  if not silent and inst.royal then
    inst.royal = nil
    inst.components.talker:Say(GetString(inst, "ANNOUNCE_KINGDESTROYED"))
    inst.sg:PushEvent("powerdown_wurt")
    inst.SoundEmitter:PlaySound("dontstarve/characters/wurt/transform_from")
  end
end
local function IsEntityDead(inst) return inst.components.health:IsDead() end
function ErodeAway(inst, erode_time)
  local time_to_erode = erode_time or 1
  local tick_time = TheSim:GetTickTime()

  if inst.DynamicShadow ~= nil then inst.DynamicShadow:Enable(false) end

  inst:StartThread(function()
    local ticks = 0
    while ticks * tick_time < time_to_erode do
      local erode_amount = ticks * tick_time / time_to_erode
      inst.AnimState:SetErosionParams(erode_amount, 0.1, 1.0)
      ticks = ticks + 1
      Yield()
    end
    inst:Remove()
  end)
end
-- changed
local WARNING_MUST_TAGS = {"monster", "hostile", "wet", "WORM_DANGER"}
-- {"tentacle"}--, "invisible"}
local function UpdateTentacleWarnings(inst)
  local disable = false -- (inst.replica.inventory ~= nil and not inst.replica.inventory:IsVisible())

  if not disable then
    local old_warnings = {}
    for t, w in pairs(inst._active_warnings) do old_warnings[t] = w end

    local x, y, z = inst.Transform:GetWorldPosition()
    local warn_dist = 15
    local tentacles = TheSim:FindEntities(x, y, z, warn_dist, WARNING_MUST_TAGS)
    for i, t in ipairs(tentacles) do
      -- change, add invisible check and tentacle check
      if t.prefab == "tentacle" and t.sg:HasStateTag("invisible") then
        local p1x, p1y, p1z = inst.Transform:GetWorldPosition()
        local p2x, p2y, p2z = t.Transform:GetWorldPosition()
        local dist = VecUtil_Length(p1x - p2x, p1z - p2z)

        if not IsEntityDead(t, true) then
          if inst._active_warnings[t] == nil then
            local fx = SpawnPrefab("wurt_tentacle_warning")
            fx.entity:SetParent(t.entity)
            inst._active_warnings[t] = fx
          else
            old_warnings[t] = nil
          end
        end
      end
    end

    for t, w in pairs(old_warnings) do
      inst._active_warnings[t] = nil
      if w:IsValid() then ErodeAway(w, 0.5) end
    end
  elseif next(inst._active_warnings) ~= nil then
    for t, w in pairs(inst._active_warnings) do if w:IsValid() then w:Remove() end end
    inst._active_warnings = {}
  end
end

local function DisableTentacleWarning(inst)
  if inst.tentacle_warning_task ~= nil then
    inst.tentacle_warning_task:Cancel()
    inst.tentacle_warning_task = nil
  end

  for t, w in pairs(inst._active_warnings) do if w:IsValid() then w:Remove() end end
  inst._active_warnings = {}
end

local function EnableTentacleWarning(inst)
  -- if inst.player_classified ~= nil then
  --    inst:ListenForEvent("playerdeactivated", DisableTentacleWarning)
  if inst.tentacle_warning_task == nil then
    inst.tentacle_warning_task = inst:DoPeriodicTask(0.1, UpdateTentacleWarnings)
  end
  -- else
  -- inst:RemoveEventCallback("playeractivated", EnableTentacleWarning)
  -- end--deleted
end

local function SetGhostMode(inst, isghost)
  -- changed
  if c_countprefabs('tentacle', true) > 0 then EnableTentacleWarning(inst) end
  do return end
  if isghost then
    DisableTentacleWarning(inst)
    inst._SetGhostMode(inst, true)
  else
    inst._SetGhostMode(inst, false)
    EnableTentacleWarning(inst)
  end
end

local function FishPreserverRate(inst, item)
  return (item ~= nil and item:HasTag("fish")) and TUNING.WURT_FISH_PRESERVER_RATE or nil
end

local function OnSave(inst, data)
  data.health_percent = inst.health_percent or inst.components.health:GetPercent()
  data.sanity_percent = inst.sanity_percent or inst.components.sanity:GetPercent()
  data.hunger_percent = inst.hunger_percent or inst.components.hunger:GetPercent()
end

local function OnPreLoad(inst, data)
  if data then
    if data.health_percent then inst.health_percent = data.health_percent end

    if data.sanity_percent then inst.sanity_percent = data.sanity_percent end

    if data.hunger_percent then inst.hunger_percent = data.hunger_percent end
  end
end

local function CLIENT_Wurt_HostileTest(inst, target)
  if target.HostileToPlayerTest ~= nil then return target:HostileToPlayerTest(inst) end
  return (target:HasTag("hostile") or target:HasTag("pig")) and not target:HasTag("merm")
           and not target:HasTag("manrabbit") and not target:HasTag("frog")
end

local function common_postinit(inst)
  inst:AddTag("playermerm")
  inst:AddTag("merm")
  inst:AddTag("mermguard")
  inst:AddTag("mermfluent")
  inst:AddTag("merm_builder")
  inst:AddTag("wet")
  inst:AddTag("stronggrip")

  inst.customidleanim = "idle_wurt"

  inst.AnimState:AddOverrideBuild("wurt_peruse")
  inst.AnimState:SetHatOffset(0, 20) -- This is not networked.

  if TheNet:GetServerGameMode() == "lavaarena" then
    -- do nothing
  elseif TheNet:GetServerGameMode() == "quagmire" then
    inst:AddTag("quagmire_shopper")
  else
    if not TheNet:IsDedicated() then
      inst._active_warnings = {}
      inst:ListenForEvent("playeractivated", EnableTentacleWarning)
    end
  end

  -- reader (from reader component) added to pristine state for optimization
  inst:AddTag("reader")

  -- aspiring_bookworm (from reader component) added to pristine state for optimization
  inst:AddTag("aspiring_bookworm")

  inst.HostileTest = CLIENT_Wurt_HostileTest
end

local function IsNonPlayerMerm(this) return this:HasTag("merm") and not this:HasTag("player") end

local MAX_TARGET_SHARES = 8
local SHARE_TARGET_DIST = 20

local function OnAttacked(inst, data)
  local attacker = data and data.attacker
  if attacker and inst.components.combat:CanTarget(attacker) then
    inst.components.combat:ShareTarget(attacker, SHARE_TARGET_DIST, IsNonPlayerMerm, MAX_TARGET_SHARES)
  end
end

local function master_postinit(inst)
  -- added
  inst.talker_path_override = "wurt/"
  inst.soundsname = "wurt"
  inst.starting_inventory = start_inv[TheNet:GetServerGameMode()] or start_inv.default

  inst:AddComponent("reader")
  -- inst.components.reader:SetAspiringBookworm(true)

  inst.components.sanity.no_moisture_penalty = true

  -- Keep in sync with merm + mermking (minus bonuses for TUNING balancing)!
  if not inst.components.foodaffinity then inst:AddComponent("foodaffinity") end -- added
  -- inst.components.foodaffinity:AddFoodtypeAffinity("VEGGIE", 1.33)
  inst.components.foodaffinity:AddPrefabAffinity("seaweed", 1.33) -- prevents the negative stats, otherwise foodtypeaffinity would have suffice
  inst.components.foodaffinity:AddPrefabAffinity("seaweed_cooked", 1.33) -- prevents the negative stats, otherwise foodtypeaffinity would have suffice
  inst.components.foodaffinity:AddPrefabAffinity("durian", 1.93) -- veggi bonus + 15
  inst.components.foodaffinity:AddPrefabAffinity("durian_cooked", 1.93) -- veggi bonus + 15

  inst:AddComponent("itemaffinity")
  inst.components.itemaffinity:AddAffinity("hutch_fishbowl", nil, TUNING.DAPPERNESS_MED, 1)
  inst.components.itemaffinity:AddAffinity(nil, "fish", TUNING.DAPPERNESS_MED, 1)
  inst.components.itemaffinity:AddAffinity(nil, "fishmeat", -TUNING.DAPPERNESS_MED_LARGE, 2)
  inst.components.itemaffinity:AddAffinity(nil, "spoiled_fish", -TUNING.DAPPERNESS_MED_LARGE, 2)

  inst:AddComponent("preserver")
  inst.components.preserver:SetPerishRateMultiplier(FishPreserverRate)

  if inst.components.eater ~= nil then
    -- inst.components.eater:SetDiet({ FOODGROUP.VEGETARIAN }, { FOODGROUP.VEGETARIAN })
    inst.components.eater.foodprefs = {"VEGGIE", "SEEDS"}
  end

  inst.components.locomotor:SetFasterOnGroundTile(WORLD_TILES.MARSH, true)

  inst.components.builder.mashturfcrafting_bonus = 2
  inst:ListenForEvent("onmermkingcreated", function() RoyalUpgrade(inst) end, TheWorld)
  inst:ListenForEvent("onmermkingdestroyed", function() RoyalDowngrade(inst) end, TheWorld)
  inst:ListenForEvent("onattacked", OnAttacked)

  inst.OnSave = OnSave
  inst.OnPreLoad = OnPreLoad

  inst:DoTaskInTime(0, function(inst)
    if TheWorld.components.mermkingmanager and TheWorld.components.mermkingmanager:HasKing() then
      RoyalUpgrade(inst)
    else
      RoyalDowngrade(inst)
    end
  end)
  -- from Island Adventure
  inst.components.itemaffinity:AddAffinity("packim_fishbone", nil, -TUNING.DAPPERNESS_MED, 2)
  inst.components.health.cantdrown_penalty = true
  inst.components.locomotor:SetFasterOnGroundTile(WORLD_TILES.TIDALMARSH, true)
  -- inst.peruse_meteor = peruse_meteor
end
local postinit = function(...)
  common_postinit(...)
  master_postinit(...)
end
return MakePlayerCharacter("wurt", prefabs, assets, postinit)
